# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ftrack_s3_accessor']

package_data = \
{'': ['*']}

install_requires = \
['boto3>=1.15.10,<2.0.0', 'ftrack-python-api>=2.0.0,<3.0.0']

extras_require = \
{'action': ['ftrack-action-handler>=0.2.1,<0.3.0']}

setup_kwargs = {
    'name': 'ftrack-s3-accessor',
    'version': '0.1.7',
    'description': 'A ftrack s3 accessor updated to work with ftrack-python-api and boto3.',
    'long_description': "# ftrack-s3-accessor\n\nAn ftrack s3 accessor updated to work with ftrack-python-api and boto3. \n\n## Requirements: \n- ftrack-python-api\n- boto3 - if used standalone \n- ftrack-action-handler (optional) - if used with the transfer components action found [here](https://bitbucket.org/!api/2.0/snippets/ftrack/B6dX/f9e89e8bf95065a6fc0541dd058863ff1ddaceb6/files/transfer_components_action.py)\n\n## Installation\n\nInstall using pip:\n    \n    pip install ftrack-s3-accessor\n\nOr with the action handler:\n\n    pip install ftrack-s3-accessor['action']\n\n## Configuration\n\nConfigure a new location within ftrack with the name 'studio.remote'. This will be used as the location for s3.\n\nCreate a storage bucket in s3 and set the bucket name using the FTRACK_S3_ACCESSOR_BUCKET environment variable (default: ftrack). Ensure your bucket name is globally unique and meets aws s3 naming restrictions.\n\nSet all other ftrack environment variables for your ftrack instance.\n\nRunning the scripts from within your environment requires you to additionally set your sources root to the accessor directory.\n\n    PYTHONPATH=./ftrack_s3_accessor\n\nEnsure you have an working aws configuration under your ~/.aws folder. You can check this by running:\n    \n    import boto3\n\nIf this fails, your aws configuration isn't setup properly. Refer to the [boto3](https://github.com/boto/boto3) documentation on how to set it up. You should only need a ~/.aws/config and ~/.aws/credentials file.\n\n## Usage\n\nThe main plugin can be found in the plugins folder. This folder may be registered using the FTRACK_EVENT_PLUGIN_PATH ftrack environment variable so that it is picked up when ftrack is started.\n\nExamples of how to use the plugin can be found in the scripts folder. The simplest way to launch ftrack with the accessor is scripts/start_ftrack_with_s3.py. \n\nIt is possible to use the [transfer components](https://bitbucket.org/!api/2.0/snippets/ftrack/B6dX/f9e89e8bf95065a6fc0541dd058863ff1ddaceb6/files/transfer_components_action.py) action to move components between local and remote storage. Ensure it is on the FTRACK_EVENT_PLUGIN_PATH (or add it to the plugins folder) and it should become available under ftracks actions menu. You will need to ensure your local storage is also correctly configured within a script when running the accessor outside of ftrack-connect, as the connect location configured by the desktop client will not be available as an option.\n",
    'author': 'Ian Wootten',
    'author_email': 'hi@niftydigits.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/niftydigits/ftrack-s3-accessor',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, !=3.5.*, !=3.6.*',
}


setup(**setup_kwargs)



import os

from io import BytesIO

BOTO_AVAILABLE = False

try:
    import boto3

    BOTO_AVAILABLE = True
except ImportError:
    pass

from ftrack_api.accessor.base import Accessor
from ftrack_api.data import FileWrapper
from ftrack_api.exception import (AccessorOperationFailedError,
                                  AccessorUnsupportedOperationError,
                                  AccessorResourceInvalidError,
                                  AccessorResourceNotFoundError,
                                  AccessorContainerNotEmptyError,
                                  AccessorParentResourceNotFoundError)


class S3File(FileWrapper):
    """S3 Buffered File.
    """

    def __init__(self, key, mode='rb'):
        """Initialise S3 file with S3 *key* and *mode*."""
        self.key = key
        self.mode = mode
        wrapped_file = BytesIO()
        super(S3File, self).__init__(wrapped_file)
        self._read()

    def flush(self):
        """Flush all changes."""
        super(S3File, self).flush()
        self._write()

    def _read(self):
        """Read all remote content from key into wrapped_file."""
        position = self.tell()
        self.seek(0)
        self.key.download_fileobj(self.wrapped_file)
        if not self.wrapped_file.closed:
            self.seek(position)

    def _write(self):
        """Write current data to remote key."""
        position = self.tell()
        self.seek(0)
        self.key.upload_fileobj(self.wrapped_file)
        if not self.wrapped_file.closed:
            self.seek(position)


class S3Accessor(Accessor):
    """Provide Amazon S3 location access.

    To use, please first install the :term:`Boto3` package.
    """

    def __init__(self, bucket_name):
        """Initialise location accessor.

        *bucket_name* is the name of the bucket to provide access to using
        *access_key* and *secretKey* credentials.
        """
        if not BOTO_AVAILABLE:
            raise NotImplementedError(
                'S3Accessor not available as failed to import boto3 package. '
                'Please install boto3 if you want to use this accessor.'
            )

        self.bucket_name = bucket_name
        self._s3 = None
        self._bucket = None
        super(S3Accessor, self).__init__()

    def __deepcopy__(self, memo):
        """Return a new S3Accessor instance

        There is a known issue on some operating systems when using :term:`Boto`
        >= 2.29 against :term:`Python` >= 2.7. The issue typically manifests
        itself with the following error::

            TypeError: cannot deepcopy this pattern object

        To work around the issue we return a new instance of the accessor which
        does not contain the compiled patterns in the Boto internals.
        """
        return S3Accessor(self.bucket_name)

    @property
    def s3(self):
        """Return S3 resource."""
        if self._s3 is None:
            self._s3 = boto3.resource('s3')

        return self._s3

    @property
    def bucket(self):
        """Return bucket."""
        if self._bucket is None:
            self._bucket = self.s3.Bucket(self.bucket_name)

        return self._bucket

    def list(self, resource_identifier):
        """Return list of entries in *resource_identifier* container.

        Each entry in the returned list should be a valid resource identifier.

        Raise :py:class:`~ftrack.ftrackerror.AccessorResourceNotFoundError` if
        *resourceIdentifier* does not exist or
        :py:class:`~ftrack.ftrackerror.AccessorResourceInvalidError` if
        *resourceIdentifier* is not a container.

        """
        if not resource_identifier.endswith('/'):
            resource_identifier += '/'

        if resource_identifier == '/':
            resource_identifier = ''

        listing = []
        for entry in self.bucket.objects.filter(Prefix=resource_identifier, Delimiter='/'):
            if entry.name != resource_identifier:
                listing.append(entry.name.rstrip('/'))

        return listing

    def exists(self, resource_identifier):
        """Return if *resourceIdentifier* is valid and exists in location."""
        # Root directory always exists
        if not resource_identifier:
            return True

        return (self.is_container(resource_identifier) or
                self.is_file(resource_identifier))

    def is_file(self, resource_identifier):
        """Return whether *resource_identifier* refers to a file."""
        # Root is a directory
        if not resource_identifier:
            return False

        resource_identifier = resource_identifier.rstrip('/')
        try:
            s3_object = self.bucket.Object(resource_identifier)
        except self.s3.meta.client.exceptions.ClientError as error:
            if error.status == 403:
                # Permission denied
                return False

            else:
                raise AccessorOperationFailedError(
                    operation='is_file',
                    resourceIdentifier=resource_identifier,
                    details=error
                )

        return s3_object.meta.data is not None

    def is_container(self, resource_identifier):
        """Return whether *resource_identifier* refers to a container."""
        # Root is a directory
        if not resource_identifier:
            return True

        # Check if list request returns any files. This avoids relying on the
        # presence of a special empty file for directories.
        if not resource_identifier.endswith('/'):
            resource_identifier += '/'

        s3_objects = self.bucket.objects.filter(Prefix=resource_identifier, Delimiter='/')

        try:
            next(iter(s3_objects))
        except StopIteration:
            return False
        else:
            return True

    def is_sequence(self, resource_identifier):
        """Return whether *resource_identifier* refers to a file sequence."""
        raise AccessorUnsupportedOperationError('is_sequence')

    def open(self, resource_identifier, mode='rb'):
        """Return :py:class:`~ftrack.Data` for *resourceIdentifier*."""
        if self.is_container(resource_identifier):
            raise AccessorResourceInvalidError(
                resource_identifier,
                message='Cannot open a directory: {resource_identifier}'
            )

        s3_object = self.bucket.Object(resource_identifier)

        if s3_object is None:
            if 'w' not in mode and 'a' not in mode:
                raise AccessorResourceNotFoundError(resource_identifier)

            if not self.is_container(self.get_container(resource_identifier)):
                raise AccessorResourceNotFoundError(
                    self.get_container(resource_identifier)
                )

            # New file
            s3_object = self.bucket.Object(resource_identifier)
            s3_object.put(Body='')

        elif 'w' in mode:
            # Truncate file
            s3_object.put(Body='')

        return S3File(s3_object, mode=mode)

    def remove(self, resource_identifier):
        """Remove *resourceIdentifier*.

        Raise :py:class:`~ftrack.ftrackerror.AccessorResourceNotFoundError` if
        *resourceIdentifier* does not exist.

        """
        if self.is_file(resource_identifier):
            self.bucket.objects.filter(Prefix=resource_identifier).delete()

        elif self.is_container(resource_identifier):
            contents = self.list(resource_identifier)
            if contents:
                raise AccessorContainerNotEmptyError(resource_identifier)

            self.bucket.objects.filter(resource_identifier + '/').delete()

        else:
            raise AccessorResourceNotFoundError(resource_identifier)

    def get_container(self, resource_identifier):
        """Return resourceIdentifier of container for *resourceIdentifier*.

        Raise
        :py:class:`~ftrack.ftrackerror.AccessorParentResourceNotFoundError` if
        container of *resourceIdentifier* could not be determined.

        """
        if os.path.normpath(resource_identifier) in ('/', ''):
            raise AccessorParentResourceNotFoundError(
                resource_identifier,
                message='Could not determine container for '
                        '{resource_identifier} as it is the root.'
            )

        return os.path.dirname(resource_identifier.rstrip('/'))

    def make_container(self, resource_identifier, recursive=True):
        """Make a container at *resourceIdentifier*.

        If *recursive* is True, also make any intermediate containers.

        """
        if not resource_identifier:
            # Root bucket directory
            return

        if not resource_identifier.endswith('/'):
            resource_identifier += '/'

        if self.exists(resource_identifier):
            if self.is_file(resource_identifier):
                raise AccessorResourceInvalidError(
                    resource_identifier,
                    message=('Resource already exists as a file: '
                             '{resourceIdentifier}')
                )

            else:
                return

        parent = self.get_container(resource_identifier)
        if not self.is_container(parent):
            if recursive:
                self.make_container(parent, recursive=recursive)
            else:
                raise AccessorParentResourceNotFoundError(parent)

        s3_object = self.bucket.Object(resource_identifier)
        s3_object.put(Body='')
